#! /home/miko/python/HydroErosion/env/bin/python3

from . utilities import mean

#  from viewLandscape import viewMap

def getDeltaHeight(map, x, y):
    deltaHeight = []
    for i in [-1, 0, 1]:
        deltaHeight.append([])
        for j in [-1, 0, 1]:
            if (
                (x + i >= len(map))
                or (x + i < 0)
                or (y + j >= len(map[0]))
                or (y + j < 0)
            ):
                # out of bounds. make sure doesn't get picked
                deltaHeight[i + 1].append(10000)
            else:
                deltaHeight[i + 1].append(map[x + i][y + j] - map[x][y])
                if abs(i + j) == 2 and i != 0:  # weight corners less
                    deltaHeight[i + 1][j + 1] *= 0.65  # 1/sqrt(2)

    return deltaHeight


# Perform actual erosion operation on map
def erodeWithDrop(map, rockmap, hydrationMap, x, y, carry):
    for time in range(25):  # drop lifespan = 25
        # find lowest surrounding point
            deltaHeight = getDeltaHeight(map, x, y)
            d_x = 0
            d_y = 0
            for i in [-1, 0, 1]:
                for j in [-1, 0, 1]:
                    if deltaHeight[i + 1][j + 1] < deltaHeight[d_x][d_y]:
                        d_x = i
                        d_y = j
            fu_x = x + d_x
            fu_y = y + d_y

            # Perform droplet move
            ch_height = map[x][y] - map[fu_x][fu_y]  # get base delta height
            rockMult = 1  # get multiplier due to bedrock
            if map[x][y] - carry * ch_height < rockmap[x][y]:
                rockMult = 0.1
                rockmap[x][y] = map[x][y] - carry * ch_height * rockMult
            map[x][y] -= carry * ch_height * rockMult
            map[fu_x][fu_y] += carry * ch_height

            x = fu_x
            y = fu_y

            # set hydrationMap
            hydrationMap[x][y] += 1


def erodeMap(heightmap, rockmap, iterate=400, carry=0.15):
    hydrationMap = [[0] * len(row) for row in heightmap]
    for i in range(iterate):
        for x in range(len(heightmap)):
            for y in range(len(heightmap[0])):
                erodeWithDrop(heightmap, rockmap, hydrationMap, x, y, carry)

    return (heightmap, hydrationMap)



import noise
from . utilities import writeCSV, normalizeNoiseMap

def generateNoiseMap(length, width, scale=190, roughness=5, shift=2.0):
    octaves = roughness
    persistence = .5
    lacunarity = shift
    map = [[noise.pnoise2(i/scale, j/scale,
        octaves=octaves,
        persistence=persistence,
        lacunarity=lacunarity,
        repeatx=length,
        repeaty=width,
        base=0
    ) for i in range(width)] for j in range(length)]
    return map

def generateTerrainMaps(x=150, y=150):
    heightmap = normalizeNoiseMap(generateNoiseMap(x,y), 0, 1)
    rockmap = normalizeNoiseMap(generateNoiseMap(x,y, roughness=8), 0, 1)
    return (heightmap, rockmap)

if __name__ == '__main__':
    heightmap, rockmap = generateTerrainMaps(300, 200)
    writeCSV(heightmap, './maps/raw/heightmap.csv')
    writeCSV(rockmap, './maps/raw/rockmap.csv')




import sys
from . utilities import readCSV, writeCSV
from . erodeLandscape import erodeMap

def main():
    args = sys.argv[1:]
    if len(args) != 4:
        print("Arguments: [source.csv] [rock.csv] [destination.csv] <#iterations>")
        quit()
    source = args[0]
    rock_source = args[1]
    destination = args[2]
    iterations = int(args[3])
    print(f"Running: ({source}, {rock_source}) #{iterations} => {destination}")
    base_map = readCSV(source)
    rock = readCSV(rock_source)
    height, hydration = erodeMap(base_map, rock, iterations)
    writeCSV(height, destination)
    writeCSV(hydration, "hydrationmap.csv")
    #  viewMap(height)

if __name__ == "__main__":
    main()



import plotly.graph_objects as go # https://plot.ly/python/3d-surface-plots/
from hydro_erosion.utilities import readCSV, normalizeNoiseMap, dim
from mayavi import mlab

# Render map
def viewMap(map):
    viewMapMayavi(map)

def viewMapPlotly(map):
    fig = go.Figure(data=[go.Surface(
        z=map,
        x=[i for i in range(len(map))],
        y=[i for i in range(len(map[0]))]
    )])
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=0))
    fig.show()

def viewMapMayavi(map, factor=40):
    # print(sum([sum(i)/len(i) for i in map])/len(map))
    map = normalizeNoiseMap(map, 0, factor)
    # map = [[j*factor for j in i] for i in map]
    # print(sum([sum(i)/len(i) for i in map])/len(map))
    mlab.surf(map)
    mlab.show()

if __name__ == '__main__':
    surf = readCSV('./maps/processed/erodedmap.csv')
    # surf = readCSV('./maps/raw/heightmap.csv')
    viewMap(surf)



# Utilities file


def normalizeNoiseMap(map, minV=0, maxV=1):
    curMinV = min([min(i) for i in map])
    curMaxV = max([max(i) for i in map])
    targetDiff = maxV - minV
    return [[ (i - curMinV) / (curMaxV - curMinV) * targetDiff + minV for i in row] for row in map]

def readCSV(filename):
    res = []
    file = open(filename)
    for line in file.readlines():
        strings = line.strip().split(',')
        res.append([float(i) for i in strings])
    file.close()
    return res

def writeCSV(data, filename):
    string = ''
    for i in data:
        for j in i:
            string += str(j) + ','
        string = string[0:-1] # remove trailing comma
        string += '\n'
    # write to file
    file = open(filename, 'w')
    file.write(string)
    file.close()

def mean(list):
    # really python. u gonna make me do this...
    return sum(list)/len(list)

def dim(a):
    if not type(a) == list:
        return []
    return [len(a)] + dim(a[0])
